# This functionality is courtesy of Anton Eskov (@antoneskov)
import datetime as dt
from abc import ABC, abstractmethod

import pandas as pd

from aeslib.data_science.utility import Utility
from aeslib.data_engineering.azure_tools import AzureKeyvaultHelper
from aeslib.data_science.interval_analysis import IntervalAnalysis, IntervalAnalysisMulti


class DataCorrectorBase(ABC):

    @abstractmethod
    def correct_dataframe(self, df: pd.DataFrame)->pd.DataFrame:
        pass

    def get_freq(self, df: pd.DataFrame)->dt.timedelta:
        return Utility.get_freq(df)


class StateRunDataCorrector(DataCorrectorBase):
    def __init__(self,
                 keyvault: AzureKeyvaultHelper,
                 df_tags: pd.DataFrame,
                 statecol_pattern: str = 'StateRun (End)',
                 interpolate: bool = True):
        self.keyvault = keyvault
        self.df_tags = df_tags
        self.statecol_pattern = statecol_pattern
        self.interpolate = interpolate

    # Assuming Time indexed dataframe
    def correct_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        assert df is not None
        assert len(df) > 2

        df_state_end = df.columns[df.columns.str.contains(self.statecol_pattern)]
        if self.interpolate:
            df_state_end.fillna(method='pad', inplace=True)
        df.update(df_state_end)

        return df

class MultiIndexDataCorrector(DataCorrectorBase):
    def __init__(self, level0: str, level1: str):
        self.level0 = level0
        self.level1 = level1

    def correct_dataframe(self, df: pd.DataFrame)->pd.DataFrame:
        res = df.drop(columns=['Unnamed: 0'])
        res.reset_index(inplace=True)
        res.set_index([self.level0, self.level1], inplace=True)
        return res


class TimeZoneDataCorrector(DataCorrectorBase):
    def __init__(self,
                 tz_source='Europe/London',
                 tz_dest='UTC',
                 ambiguous='infer',
                 neutralize_tz: bool = True,
                 level_name='TimeStamp'):
        self.tz_source = tz_source
        self.ambiguous = ambiguous
        self.tz_dest = tz_dest
        self.neutralize_tz = neutralize_tz
        self.level_name = level_name

    def correct_dataframe(self, df: pd.DataFrame)->pd.DataFrame:
        level = self.level_name if isinstance(df.index, pd.MultiIndex) else None
        res = df.tz_localize(self.tz_source,
                             ambiguous=self.ambiguous,
                             copy=False,
                             level=level).tz_convert(self.tz_dest, level=level)
        if self.neutralize_tz:
            res = res.tz_convert(None, level=level)
        res.sort_index(inplace=True)
        return res
    
# Assumption:
#  data Index - DateTime, timezone neutral, contains UTC timestamps (no duplicates)
#  cols_to_interpolate does not include "step columns"
class InterpolateDropDataCorrector(DataCorrectorBase):
    def __init__(self, drop_threshold: dt.timedelta, cols_to_interpolate=None, method='linear'):
        self.drop_threshold = drop_threshold
        self.cols_to_interpolate = cols_to_interpolate
        self.method = method
        self.all_intervals = None
        self.dropped_intervals = None


    def get_empty_intervals(self, df: pd.DataFrame)-> pd.DataFrame:
        cols = df.columns if self.cols_to_interpolate is None else self.cols_to_interpolate
        return IntervalAnalysis.get_empty_intervals(df, cols)


    def correct_dataframe(self, df: pd.DataFrame)->pd.DataFrame:
        cols_to_interpolate = \
            df.columns if self.cols_to_interpolate is None else self.cols_to_interpolate

        # Prepare list of intervals
        df_interval = self.get_empty_intervals(df)

        # Drop big intervals
        df_drops = df_interval[df_interval.duration > self.drop_threshold]
        dfres = df

        for drop in df_drops.itertuples():
            dfres = dfres.drop(axis=0, index=dfres.loc[(dfres.index > drop.interval.left) \
                & (dfres.index < drop.interval.right)].index)

        # Interpolate
        dfres[cols_to_interpolate] = dfres[cols_to_interpolate].interpolate(method=self.method,
                                                                            axis=0)

        self.all_intervals = df_interval
        self.dropped_intervals = df_drops
        return dfres

# Assumption:
#  data Index - DateTime, timezone neutral, contains UTC timestamps (no duplicates)
#  cols_to_interpolate does not include "step columns"
class InterpolateDropMultiDataCorrector(InterpolateDropDataCorrector):
    # def __init__(self, drop_threshold: dt.timedelta, cols_to_interpolate = None, method='linear'):
    #     self.drop_threshold = drop_threshold
    #     self.cols_to_interpolate = cols_to_interpolate
    #     self.method = method
    #     self.all_intervals = None
    #     self.dropped_intervals = None


    def get_empty_intervals(self, df: pd.DataFrame) -> pd.DataFrame:
        cols = df.columns if self.cols_to_interpolate is None else self.cols_to_interpolate
        return IntervalAnalysisMulti.get_empty_intervals(df, cols)


    def correct_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        slice_levels = IntervalAnalysisMulti.get_slice_levels(df)
        timeindex_name = IntervalAnalysisMulti.get_timeindex_name(df)
        dfres = None
        for slicer, df_slice in IntervalAnalysisMulti.iterate_time_levels(df):
            df_corrected = super().correct_dataframe(df_slice)
            df_corrected.reset_index(inplace=True)

            if len(slice_levels) == 1:
                slicer = [slicer]
        
            for idx, level in enumerate(slice_levels):
                df_corrected[level] = slicer[idx]

            dfres = df_corrected if dfres is None else pd.concat([dfres, df_corrected], axis=0)


        dfres.set_index(slice_levels + [timeindex_name], inplace=True)
        dfres.sort_index(level=dfres.index.names, inplace=True)
        return dfres


class DataCorrector:
    @staticmethod
    def execute(df: pd.DataFrame, df_tagids: pd.DataFrame, interpolate: bool = True) ->pd.DataFrame:
        keyvault = AzureKeyvaultHelper(key_vault_name='arnts-keyvault')
        sr_corrector = StateRunDataCorrector(keyvault=keyvault,
                                             df_tags=df_tagids,
                                             interpolate=interpolate)
        # Add more correctors and execute these in a sequence, if needed
        return sr_corrector.correct_dataframe(df)
        


import datetime as dt
import pandas as pd
import numpy as np

class Utility:

    @staticmethod
    def get_freq(df: pd.DataFrame) -> dt.timedelta:
        assert len(df) > 2, "Not enough data in the data set"

        check_level = None

        #Compensate for MultiIndex - find Datetime level
        if isinstance(df.index, pd.MultiIndex):
            for level in df.index.levels:
                if isinstance(level, pd.DatetimeIndex):
                    check_level = level
                    break
        elif isinstance(df.index, pd.DatetimeIndex):
            check_level = df.index

        assert check_level is not None, "Cannot define frequency - no DatetimeIndex given"
        return check_level[1]-check_level[0]

    @staticmethod
    def mean_absolute_percentage_error(y_true, y_pred):
        y_true, y_pred = np.array(y_true), np.array(y_pred)
        return np.mean(np.abs((y_true - y_pred) / y_true)) * 100



# This functionality is courtesy of Thibaut Forest (@thif)
import datetime as dt
import json
import re
import requests
import pandas as pd
import numpy as np
import urllib3

from requests.auth import HTTPBasicAuth

from aeslib.data_engineering import azure_tools as at

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

class BazefieldClient:
    def __init__(self,
                 keyvault: at.AzureKeyvaultHelper,
                 from_timestamp: dt.datetime,
                 to_timestamp: dt.datetime,
                 reg_ex_strings: list,
                 aggregates: str = 'Raw',#'Min,Max,Average,End,Standard deviation',
                 interval: str = '0', #'10',
                 destination: str = '.',
                 datetime_format: str = 'dd-MM-yyyy HH:mm:ss.fff',
                 calender_unit: str = 'Second',
                 use_asset_title: bool = False,
                 use_interval: bool = True,
                 export_in_utc: bool = True):
        self.keyvault = keyvault
        self.from_timestamp = from_timestamp
        self.to_timestamp = to_timestamp
        self.reg_ex_strings = reg_ex_strings
        self.aggregates = aggregates
        self.interval = interval
        self.destination = destination
        self.datetime_format = datetime_format
        self.calender_unit = calender_unit
        self.use_asset_title = use_asset_title
        self.use_interval = use_interval
        self.export_in_utc = export_in_utc
        
    @staticmethod
    def _get_datetime_as_string(timestamp: dt.datetime.timestamp,
                                string_format="%Y_%m_%d_%H_%M_%S") -> str:
        time_stamp_as_datetime = dt.datetime.fromtimestamp(timestamp / 1000)
        return time_stamp_as_datetime.strftime(string_format)

    @staticmethod
    def _reduce_dataframe_size(df: pd.DataFrame) -> pd.DataFrame:
        df.loc[:, df.select_dtypes(include=np.float64).columns] \
            = df.select_dtypes(include=np.float64).astype(np.float32)
        return df

    @staticmethod
    def _get_bazefield_headers() -> dict:
        headers = {'Content-Type': 'application/json;charset=UTF-8',
                   'Accept': 'application/json, text/plain, */*'}
        return headers

    @staticmethod
    def _match_taglist_to_regular_expression(tag_list: list, regular_expression: str) -> list:
        tag_ids = [t["tagId"] for t in tag_list
                   if re.match(regular_expression, t["tagName"])]
        return tag_ids

    def _get_bazefield_authentication(self) -> HTTPBasicAuth:
        return HTTPBasicAuth(self.keyvault.get_secret(secret_name='bazefield-api-key'), '')

    def get_full_tag_list(self):
        auth = self._get_bazefield_authentication()
        tag_list_url = self.keyvault.get_secret(secret_name='bazefield-tagid-url')
        tag_list = requests.get(url=tag_list_url, verify=False, auth=auth)
        return json.loads(tag_list.text)

    def get_transformed_tag_list_from_bazefield(self, reg_ex_strings: list):
        tag_list = self.get_full_tag_list()

        tag_ids = []

        for reg_ex in reg_ex_strings:
            tag_ids = tag_ids + self._match_taglist_to_regular_expression(tag_list=tag_list,
                                                                          regular_expression=reg_ex)

        tag_list_to_download = {"tagIds": str(tag_ids)[1:-1].replace(" ", "")}

        tag_list_to_download["dateTimeFormat"] = self.datetime_format
        if self.calender_unit:
            tag_list_to_download["calenderUnit"] = self.calender_unit
        #tag_list_to_download["useAssetTitle"] = self.use_asset_title
        if self.use_interval:
            tag_list_to_download["useInterval"] = self.use_interval
        tag_list_to_download["exportInUtc"] = self.export_in_utc
        return json.dumps(tag_list_to_download)

    def prepare_download(self,
                         from_timestamp: int,
                         aggregates: str,
                         interval: str,
                         reg_ex_strings: list,
                         download_file_resolution=1000*3600*24):

        from_timestamp_as_string = self._get_datetime_as_string(from_timestamp)
        end_timestamp = from_timestamp + download_file_resolution

        print("Downloading " + from_timestamp_as_string)

        url = self.keyvault.get_secret(secret_name='bazefield-data-export-url')
        auth = self._get_bazefield_authentication()
        headers = self._get_bazefield_headers()
        proxy = {'https': self.keyvault.get_secret(secret_name='equinor-proxy')}
        data = self.get_transformed_tag_list_from_bazefield(reg_ex_strings)

        res = requests.post(url=url.format(from_timestamp, end_timestamp, interval, aggregates),
                            auth=auth,
                            headers=headers,
                            verify=False,
                            proxies=proxy,
                            data=data)

        filename = res.text
        next_timestamp = end_timestamp

        return filename, next_timestamp, from_timestamp_as_string

    def download_file(self, filename: str, from_timestamp_as_string: str):
        csv_name = from_timestamp_as_string + '.csv'
        filename = filename.replace(".txt", "")

        url = self.keyvault.get_secret(secret_name='bazefield-get-file-url')
        auth = self._get_bazefield_authentication()
        proxy = {'https': self.keyvault.get_secret(secret_name='equinor-proxy')}

        res = requests.get(url=url.format(filename),
                           auth=auth,
                           verify=False,
                           proxies=proxy)

        file = res.content.decode()

        file_path = self.destination + csv_name

        with open(file_path, 'w') as f:
            f.write(file)
            print("Finished.")
            print("Downloaded to "+ str(file_path))

        df = pd.read_csv(file_path, sep=';')
        df = self._reduce_dataframe_size(df)

        df.to_csv(file_path, sep=';')

        return True


    def download_data_from_bazefield_as_csv(self):
        """Downloads data from Bazefield using the Bazefield API.
        
        Arguments:
            from_timestamp {dt.datetime} -- Download data from and including this time.
            to_timestamp {dt.datetime} -- Download data to and including this time.
            aggregates {str} -- Which aggregates to download as specified by Bazefield.
            interval {str} -- Data interval to download. Example: '10Min'.
            reg_ex_strings {list} -- A list of regular expressions to match tags against.
        """
        from_timestamp_int = int(self.from_timestamp.timestamp()) * 1000
        to_timestamp_int = int(self.to_timestamp.timestamp())

        while True:
            try:
                filename, next_timestamp, from_timestamp_as_string = \
                    self.prepare_download(from_timestamp=from_timestamp_int,
                                          aggregates=self.aggregates,
                                          interval=self.interval,
                                          reg_ex_strings=self.reg_ex_strings)
                print('Downloading ' + from_timestamp_as_string)
                print(filename)
                _ = self.download_file(filename, from_timestamp_as_string)
            except Exception as e:
                print(e)
                from_timestamp_int = next_timestamp
                continue
            
            if dt.datetime.fromtimestamp(next_timestamp / 1000) \
            > dt.datetime.fromtimestamp(to_timestamp_int):
                return
            else:
                from_timestamp_int = next_timestamp



from azure.keyvault.secrets import SecretClient
from azure.common.client_factory import get_client_from_cli_profile
from azure.mgmt.compute import ComputeManagementClient

class AzureKeyvaultHelper:
    def __init__(self, key_vault_name: str):
        self.key_vault_uri = "https://" + key_vault_name + ".vault.azure.net"

    def get_secret(self, secret_name: str):
        credential = get_client_from_cli_profile(ComputeManagementClient).config.credentials
        client = SecretClient(self.key_vault_uri, credential)
        return client.get_secret(secret_name).value

    def set_secret(self, secret_name: str, secret_value):
        credential = get_client_from_cli_profile(ComputeManagementClient).config.credentials
        client = SecretClient(self.key_vault_uri, credential)
        client.set_secret(secret_name, secret_value)
